#ifndef SRC_CLCS_EVALUATION_H
#define SRC_CLCS_EVALUATION_H

#include "CLCS_solution.h"
#include "CLCS_inst.h"

using namespace clcs;

namespace clcs_helper {
        /** helper functions to calculate EX guidance **/
	long double stable_pow(double v, int sigma, double power, int step);
	long double Approx_term_k(double l, int Sigma, long double pi);
	int l_min(const Node &v, const CLCS_inst *inst);
	long double term_k(const Node &v, const CLCS_inst *inst, int k);
	long double fastEX(const Node &v, const CLCS_inst *inst, int bp, int ep);
}

namespace clcs_eval {

	/** type definition for upper bound functions */
	typedef int (*upperBound)(const Node &v, const CLCS_inst *inst);
	/** type definition for heuristic functions */
	typedef long double (*heuristic)(const Node &v, const CLCS_inst *inst, int p);

	/** calculate greedy criterion */
	double greedy(const Node &v, const CLCS_inst *inst);
	/** H_prob heuristic **/
	long double h_prob(const Node &v, const CLCS_inst *inst, int p);
	/** min-avg heuristic **/
	long double h_min_avg(const Node &v, const CLCS_inst *inst, int p);
	/** k-norm for tie-breaking in min-avg heuristic **/
	long double k_norm(const Node &v, const CLCS_inst *inst, int p);
	/** pow-based heuristic */
	long double h_power(const Node &v, const CLCS_inst *inst, int p);
	/** expected length calculation heuristic */
	long double ex(const Node &v, const CLCS_inst *inst, int p);
	/** Upper bound for the LCSP **/
	int ub(const Node &v, const CLCS_inst *inst);
}

#endif //SRC_CLCS_EVALUATION_H
