#ifndef CLCS_SOLUTION_H
#define CLCS_SOLUTION_H

#include <string>
#include <vector>
#include <math.h>
#include <algorithm>
#include <climits>
#include <memory>
#include <mh_util.h>
#include <mh_solution.h>
#include <mh_param.h>
#include "CLCS_inst.h"
#include <vector>
#include <queue>
#include <unordered_map>
#include <unordered_set>

using namespace std;
namespace clcs {

extern mh::int_param algorithm;
extern mh::double_param time;

/** A Node structure definition */
struct Node {
	shared_ptr<Node> parent;
	vector<int16_t> pL;
	int16_t l_v = 0;
	int16_t u_v = 0;

	Node() {
	}
        void print()
        {
            cout << " l_v: " << l_v << " " << " u_v: " << u_v << endl;
            for (int i = 0; i < (int) pL.size(); ++i) 
                 cout << pL[ i ] << endl;
        }

	~Node() {
		pL.clear();
	}
};

/** Structure used for the Illinopulos and Rahman algorithm  */
struct IR_Item {
	int val = 0;
	int x = 0;
	int y = 0;
	int z = 0;

	IR_Item() = default;
	IR_Item(int val, int x, int y, int z) : val(val), x(x), y(y), z(z) {}
};

/** Structure used for the Deorowicz algorithm */
struct Deo_Item {
	int i = INT_MIN;
	int j = INT_MIN;
	int len = INT_MIN;

	Deo_Item() = default;
	Deo_Item(int i, int j, int len) : i(i), j(j), len(len) {}
};

/** Structure used for the Diagonal approach */
struct Btriple {

public:
    int x;
    int y;
    int z;
public:
    Btriple(): x(-1), y(-1), z(-1){}; 
    Btriple(int posA, int posB, int posT) { x = posA; y = posB; z = posT; }

public:    
     
    int xx() const
    {
       return x;
    }
    int yy() const
    {
       return y;
    }
    int zz() const
    {
       return z;
    }
    bool operator < (Btriple &c) 
    {
         if( c.zz() != zz())
             return c.zz() < zz();
         else
         if( c.yy() != yy())
             return c.yy() < yy();
         else
         if( c.xx() != xx())
             return c.xx() < xx();
         else 
             return false;

    }
    
    bool operator == (Btriple& c)
    {
         return (c.xx() == xx() and c.yy() == yy() and c.zz() == zz());
    }
    /*Btriple& operator=(Btriple& c1)
    {
          x = c1.x; y = c1.y; z = c1.z;
          return (*this);
    }*/

};
/** Custom Hash Functor that will compute the hash on the
    passed string objects length (1000) for instances with larger m; we should increase 1000 to a larger number
*/
struct BtripleHasher
{
  size_t operator()(const Btriple & obj) const
  {
       return 1000 * 1000 * obj.xx() + 1000 * obj.yy()  + obj.zz();
  }
};

// Custom comparator that compares the string objects by length
struct BtripleComparator
{
  bool operator()(const Btriple & obj1, const Btriple& obj2) const
  {
       if (obj1.xx() == obj2.xx() and obj1.yy() == obj2.yy() and obj1.zz() == obj2.zz())
           return true;
       return false;
  }
};



/** Solution class for CLCS problem  */

class CLCS_solution : public mh::mh_solution {

public:
	const CLCS_inst *inst; ///< A pointer to the respective problem instance.
	std::vector<int16_t> s;///< CLCS solution vector s (store solution)
	int UB;  // proven upper bound
	/** statistics **/
	int iter = 0; // follow the number of BS or ACS iterations
	double timeStart = 0.0;
	double tbest = 0;  ///< time when best solution was found
	int iter_best = 0; ///< number of iteration when best solution found (not used here, planed for the future anytime search variants)
	double gap = 100.0;///< needed for the exact search 

	/** Constructor creating an empty solution. */
	CLCS_solution(const CLCS_inst *inst) : mh::mh_solution(1, ""), inst(inst), UB(inst->UB) {
		s.reserve(inst->UB);
	}
	/** Constructor for creating a partial solution from the given array and middle letter information. */
	CLCS_solution(const CLCS_inst *inst, std::vector<int16_t> &s, int UB)
			: mh::mh_solution(1, ""), inst(inst), s(s), UB(UB) {
		/** Dynamically cast a const mh::mh_solution reference to a reference of this class. */
	}
	/** Copy constructor. */
	CLCS_solution(const CLCS_solution &sol) : mh::mh_solution(1, ""), inst(sol.inst),s(sol.s), UB(sol.UB)  {
	}
	/** Destructor. */
	virtual ~CLCS_solution() {
	}
	/** Creates an uninitialized/empty solution. */

	mh::mh_solution *createUninitialized() const override {
		return new CLCS_solution(inst);
	}

	/** Clone this solution. */
	mh::mh_solution *clone() const override { return new CLCS_solution(*this); }

	/** Dynamically cast an mh::mh_solution reference to a reference of this class. */
	static CLCS_solution &cast(mh::mh_solution &ref) {
		return (dynamic_cast<CLCS_solution &>(ref));
	}

	static const CLCS_solution &cast(const mh::mh_solution &ref) {
		return (dynamic_cast<const CLCS_solution &>(ref));
	}

	/** Copy a solution. */
	void copy(const mh_solution &sol) override;

	/** Clears the solution. */
	void clear() {
		UB = inst->UB;
		s.clear();
		invalidate(); // invalidate --> empty solution
	}

	/** Returns the complete solution as a string. */
	std::string getSolution();

	/** Initializes the solution with the given string LCS. */
	void decodeSolution(const std::string &LCS);

	/** The objective function corresponds to the length of the solution. */
	double objective() override {
		return s.size();
	}

	CLCS_solution &operator=(const CLCS_solution &solution);

	/** Writes out the solution in human readable form. */
	virtual void write(std::ostream &os, int detailed = 0);

	/** Saves a solution to a file. Does nothing if fname==NULL. */
	void save(const std::string &fname) override;

	/** Loads a solution from a file. */
	void load(const std::string &fname) override;

	/** Approximation procedure for the CLCS */
	void Approx();

	/** DP for 2-CLCSP problem: derived from Chin et al. **/
	int DP_Chin();

	/** 2-CLCS and DP-approaches: AE approach **/
	int DP_AE();

	/** 2-CLCS and DP-approaches: Iliopoulos-Rahman approach **/
	int DP_IR();

	/** 2-CLCS and DP-approaches: Deorowicz **/
	int DP_Deo();

	/** BS procedure for the CLCS */
	void BS();

	/** A* search for the CLCS */
	void A_Star();

	/** Greedy procedure for the CLCS */
	void Greedy();

	/** Greedy procedure for the CLCS */
	void RandomizedGreedy();

	/** check if the algorithm has to b terminated due to exceeding time limitation */
	bool terminate();
	/** Checking the feasibility of solution */
	void checkFeasibility();
	/** Get all feasible letters for a certain node */
	map<int, vector<int16_t>> findFeasibleSigma(const Node &v) const;
	/** Get all feasible, non-dominated letters for a certain node */
	map<int, vector<int16_t>> findSigmaNd(const Node &v) const;
	/** Check if position vector pL_1 is dominated by position vector pL_2 */
	bool check_domination(const vector<int16_t> &pL_1, const vector<int16_t> &pL_2) const;

	/** return a new child by extending parent node by one letter */
	shared_ptr<Node> expandNode(const shared_ptr<Node> &parent, int letter, const vector<int16_t> &pL) const;

	/** derive Solution from leaf node */
	vector<int16_t> deriveSolution(Node *v) const;
        /** Diagonal approach  **/
        int  Diagonal_based();
        /** helper functions for Diagonal-based approach (for the two following function see the corresponding paper and the pseudocode of the method */
        void Expand(Btriple& c, vector<vector<int>>& NextB, vector<int>& limitB, int i, int Tlen, vector<Btriple>& D_il);
        int  Diagonal(vector< int> T, vector< int> A, vector< int>B, int Tlen, int size);
        void Domination(vector<Btriple>& set1, vector<Btriple>& set2, int Tlen);
        /* check domination in Hung et al. approach */
        bool dominates(Btriple& c1, Btriple& c2);
        void display(vector<Btriple>& D);
        /* retrieve the solution form Hung et al. approach */
        void retrieve_diagonal(vector<vector<vector<Btriple>>>& D, int L);
};

} // clcs namespace

#endif /* CLCS_SOLUTION_H */
